/* vifm
 * Copyright (C) 2001 Ken Steen.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define _GNU_SOURCE /* I don't know how portable this is but it is
					   needed in Linux for the ncurses wide char
					   functions
					   */

#include <grp.h> /* getgrgid() */
#include <signal.h> /* signal() */
#include <stdlib.h> /* malloc */
#include <sys/stat.h> /* stat */
#include <dirent.h> /* DIR */
#include <pwd.h> /* getpwent() */
#include <string.h>
#include <time.h>
#include <termios.h> /* struct winsize */
#include <sys/ioctl.h>

#include "../config.h"

#include "color_scheme.h"
#include "config.h" /* for menu colors */
#include "file_info.h"
#include "filelist.h"
#include "macros.h"
#include "menus.h"
#include "opt_handlers.h"
#include "signals.h"
#include "status.h"
#include "ui.h"
#include "utils.h"

static void _gnuc_noreturn
finish(const char *message)
{
	endwin();
	write_config_file();
	write_info_file();
	printf("%s", message);
	exit(0);
}

void
update_pos_window(FileView *view)
{
	char buf[13];

	if(curr_stats.freeze)
		return;
	werase(pos_win);
	snprintf(buf, sizeof(buf), "%d-%d ", view->list_pos + 1, view->list_rows);
	mvwaddstr(pos_win, 0, 13 - strlen(buf), buf);
	wnoutrefresh(pos_win);
}

static void
get_id_string(FileView *view, size_t len, char *buf)
{
	char uid_buf[26];
	char gid_buf[26];
	struct passwd *pwd_buf;
	struct group *group_buf;

	if((pwd_buf = getpwuid(view->dir_entry[view->list_pos].uid)) == NULL)
	{
		snprintf(uid_buf, sizeof(uid_buf), "%d",
				(int) view->dir_entry[view->list_pos].uid);
	}
	else
	{
		snprintf(uid_buf, sizeof(uid_buf), "%s", pwd_buf->pw_name);
	}

	if((group_buf = getgrgid(view->dir_entry[view->list_pos].gid)) == NULL)
	{
		snprintf(gid_buf, sizeof(gid_buf), "%d",
				(int) view->dir_entry[view->list_pos].gid);
	}
	else
	{
		snprintf(gid_buf, sizeof(gid_buf), "%s", group_buf->gr_name);
	}

	snprintf(buf, len, "  %s:%s", uid_buf, gid_buf);
}

void
update_stat_window(FileView *view)
{
	char name_buf[160*2 + 1];
	char perm_buf[26];
	char size_buf[56];
	char id_buf[52];
	int x, y;
	int cur_x;
	size_t print_width;
	char *current_file;

	getmaxyx(stat_win, y, x);
	current_file = get_current_file_name(view);
	print_width = get_real_string_width(current_file, 20 + MAX(0, x - 83));
	snprintf(name_buf, MIN(sizeof(name_buf), print_width + 1), "%s",
			current_file);
	friendly_size_notation(view->dir_entry[view->list_pos].size,
			sizeof(size_buf), size_buf);

	get_id_string(view, sizeof(id_buf), id_buf);
	get_perm_string(perm_buf, sizeof(perm_buf),
			view->dir_entry[view->list_pos].mode);

	werase(stat_win);
	cur_x = 2;
	mvwaddstr(stat_win, 0, cur_x, name_buf);
	cur_x += 22;
	if(x > 83)
		cur_x += x - 83;
	mvwaddstr(stat_win, 0, cur_x, size_buf);
	cur_x += 12;
	mvwaddstr(stat_win, 0, cur_x, perm_buf);
	cur_x += 10;

	snprintf(name_buf, sizeof(name_buf), "%d %s filtered",
			view->filtered, view->filtered == 1 ? "file" : "files");
	if(view->filtered > 0)
		mvwaddstr(stat_win, 0, x - (strlen(name_buf) + 2) , name_buf);

	if(cur_x + strlen(id_buf) + 1 > x - (strlen(name_buf) + 2))
		*strchr(id_buf, ':') = '\0';
	if(cur_x + strlen(id_buf) + 1 > x - (strlen(name_buf) + 2))
		id_buf[0] = '\0';
	mvwaddstr(stat_win, 0, cur_x, id_buf);

	wnoutrefresh(stat_win);
}

/*
 * Repeats last message if message is NULL
 */
void
status_bar_message(const char *message)
{
	static char msg[512];

	if(!curr_stats.vifm_started)
		return;

	if(message != NULL)
		snprintf(msg, sizeof(msg), "%s", message);

	werase(status_bar);
	wmove(status_bar, 0, 0);
	wprintw(status_bar, "%s", msg);
	wnoutrefresh(status_bar);
}

int
setup_ncurses_interface(void)
{
	int screen_x, screen_y;
	int x, y;
	int i;
	int color_scheme;

	initscr();
	noecho();
	nonl();
	raw();

	curs_set(0);

	getmaxyx(stdscr, screen_y, screen_x);
	/* screen is too small to be useful*/
	if(screen_y < 10)
		finish("Terminal is too small to run vifm\n");
	if(screen_x < 30)
		finish("Terminal is too small to run vifm\n");

	if(!has_colors())
		finish("Vifm requires a console that can support color.\n");

	start_color();
	/* Changed for pdcurses */
	use_default_colors();

	for(i = 0; i < cfg.color_scheme_num; i++)
	{
		int x;
		for(x = 0; x < MAXNUM_COLOR; x++)
			init_pair(col_schemes[i].color[x].name, col_schemes[i].color[x].fg,
					col_schemes[i].color[x].bg);
	}

	color_scheme = cfg.color_scheme_cur*MAXNUM_COLOR;
	lwin.color_scheme = color_scheme;
	rwin.color_scheme = color_scheme;

	werase(stdscr);

	menu_win = newwin(screen_y - 1, screen_x, 0, 0);
	wbkgdset(menu_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	werase(menu_win);

	sort_win = newwin(NUM_SORT_OPTIONS + 3, 30, (screen_y -12)/2, (screen_x -30)/2);
	wbkgdset(sort_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	werase(sort_win);

	change_win = newwin(20, 30, (screen_y -20)/2, (screen_x -30)/2);
	wbkgdset(change_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	werase(change_win);

	error_win = newwin(10, screen_x -2, (screen_y -10)/2, 1);
	wbkgdset(error_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	werase(error_win);

	lborder = newwin(screen_y - 2, 1, 0, 0);

	wbkgdset(lborder, COLOR_PAIR(color_scheme + BORDER_COLOR));

	werase(lborder);

	if(curr_stats.number_of_windows == 1)
		lwin.title = newwin(0, screen_x - 2, 0, 1);
	else
		lwin.title = newwin(1, screen_x/2 - 2, 0, 1);

	wattrset(lwin.title, A_BOLD);
	wbkgdset(lwin.title, COLOR_PAIR(color_scheme + BORDER_COLOR));

	werase(lwin.title);

	if(curr_stats.number_of_windows == 1)
		lwin.win = newwin(screen_y - 3, screen_x - 2, 1, 1);
	else
		lwin.win = newwin(screen_y - 3, screen_x/2 - 2, 1, 1);

	wbkgdset(lwin.win, COLOR_PAIR(color_scheme + WIN_COLOR));
	wattrset(lwin.win, A_BOLD);
	wattron(lwin.win, A_BOLD);
	werase(lwin.win);
	getmaxyx(lwin.win, y, x);
	lwin.window_rows = y -1;
	lwin.window_width = x -1;

	mborder = newwin(screen_y, 2, 0, screen_x/2 - 1);

	wbkgdset(mborder, COLOR_PAIR(color_scheme + BORDER_COLOR));

	werase(mborder);

	if(curr_stats.number_of_windows == 1)
		rwin.title = newwin(0, screen_x - 2, 0, 1);
	else
		rwin.title = newwin(1, screen_x/2 - 1, 0, screen_x/2 + 1);

	wbkgdset(rwin.title, COLOR_PAIR(color_scheme + BORDER_COLOR));
	wattrset(rwin.title, A_BOLD);
	wattroff(rwin.title, A_BOLD);

	werase(rwin.title);

	if(curr_stats.number_of_windows == 1)
		rwin.win = newwin(screen_y - 3, screen_x -2 , 1, 1);
	else
		rwin.win = newwin(screen_y - 3, screen_x/2 -2 , 1, screen_x/2 +1);

	wattrset(rwin.win, A_BOLD);
	wattron(rwin.win, A_BOLD);
	wbkgdset(rwin.win, COLOR_PAIR(color_scheme + WIN_COLOR));
	werase(rwin.win);
	getmaxyx(rwin.win, y, x);
	rwin.window_rows = y - 1;
	rwin.window_width = x -1;

	if(screen_x % 2)
		rborder = newwin(screen_y - 2, 2, 0, screen_x -2);
	else
		rborder = newwin(screen_y - 2, 1, 0, screen_x -1);

	wbkgdset(rborder, COLOR_PAIR(color_scheme + BORDER_COLOR));

	werase(rborder);

	stat_win = newwin(1, screen_x, screen_y -2, 0);

	wbkgdset(stat_win, COLOR_PAIR(color_scheme + BORDER_COLOR));

	werase(stat_win);

	status_bar = newwin(1, screen_x - 19, screen_y -1, 0);
#ifdef ENABLE_EXTENDED_KEYS
	keypad(status_bar, TRUE);
#endif /* ENABLE_EXTENDED_KEYS */
	wattrset(status_bar, A_BOLD);
	wattron(status_bar, A_BOLD);
	wbkgdset(status_bar, COLOR_PAIR(color_scheme + STATUS_BAR_COLOR));
	werase(status_bar);

	pos_win = newwin(1, 13, screen_y - 1, screen_x - 13);
	wattrset(pos_win, A_BOLD);
	wattron(pos_win, A_BOLD);
	wbkgdset(pos_win, COLOR_PAIR(color_scheme + STATUS_BAR_COLOR));
	werase(pos_win);

	input_win = newwin(1, 6, screen_y - 1, screen_x -19);
	wattrset(input_win, A_BOLD);
	wattron(input_win, A_BOLD);
	wbkgdset(input_win, COLOR_PAIR(color_scheme + STATUS_BAR_COLOR));
	werase(input_win);

	wnoutrefresh(lwin.title);
	wnoutrefresh(lwin.win);
	wnoutrefresh(rwin.win);
	wnoutrefresh(rwin.title);
	wnoutrefresh(stat_win);
	wnoutrefresh(pos_win);
	wnoutrefresh(input_win);
	wnoutrefresh(lborder);
	wnoutrefresh(mborder);
	wnoutrefresh(rborder);
	wnoutrefresh(status_bar);

	return 1;
}

void
is_term_working(void)
{
	struct winsize ws;
	if(ioctl(0, TIOCGWINSZ, &ws) == -1)
		finish("Terminal error");
	if(ws.ws_row < 10)
		finish("Terminal is too small to run vifm\n");
	if(ws.ws_col < 30)
		finish("Terminal is too small to run vifm\n");
}

static void
resize_window(void)
{
	int screen_x, screen_y;
	int x, y;
	struct winsize ws;

	curr_stats.freeze = 1;

	ioctl(0, TIOCGWINSZ, &ws);

	/* changed for pdcurses */
	resize_term(ws.ws_row, ws.ws_col);

	getmaxyx(stdscr, screen_y, screen_x);

	if(screen_y < 10)
		finish("Terminal is too small to run vifm\n");
	if(screen_x < 30)
		finish("Terminal is too small to run vifm\n");

	wclear(stdscr);
	wclear(lwin.title);
	wclear(lwin.win);
	wclear(rwin.title);
	wclear(rwin.win);
	wclear(stat_win);
	wclear(status_bar);
	wclear(pos_win);
	wclear(input_win);
	wclear(rborder);
	wclear(mborder);
	wclear(lborder);

	wresize(stdscr, screen_y, screen_x);
	wresize(menu_win, screen_y - 1, screen_x);
	wresize(error_win, (screen_y -10)/2, screen_x -2);
	mvwin(error_win, (screen_y -10)/2, 1);
	wresize(lborder, screen_y -2, 1);

	if(curr_stats.number_of_windows == 1)
	{
		wresize(lwin.title, 1, screen_x -1);
		wresize(lwin.win, screen_y -3, screen_x -2);
		getmaxyx(lwin.win, y, x);
		lwin.window_width = x -1;
		lwin.window_rows = y -1;

		wresize(rwin.title, 1, screen_x -1);
		mvwin(rwin.title, 0, 1);
		wresize(rwin.win, screen_y -3, screen_x -2);
		mvwin(rwin.win, 1, 1);
		getmaxyx(rwin.win, y, x);
		rwin.window_width = x -1;
		rwin.window_rows = y -1;
	}
	else
	{
		wresize(lwin.title, 1, screen_x/2 -2);
		wresize(lwin.win, screen_y -3, screen_x/2 -2);
		getmaxyx(lwin.win, y, x);
		lwin.window_width = x -1;
		lwin.window_rows = y -1;

		mvwin(mborder, 0, screen_x/2 -1);
		wresize(mborder, screen_y, 2);

		wresize(rwin.title, 1, screen_x/2 -2);
		mvwin(rwin.title, 0, screen_x/2 +1);

		wresize(rwin.win, screen_y -3, screen_x/2 -2);
		mvwin(rwin.win, 1, screen_x/2 +1);
		getmaxyx(rwin.win, y, x);
		rwin.window_width = x -1;
		rwin.window_rows = y -1;
	}

	if(screen_x % 2)
	{
		wresize(rborder, screen_y - 2, 2);
		mvwin(rborder, 0, screen_x - 2);
	}
	else
	{
		wresize(rborder, screen_y - 2, 1);
		mvwin(rborder, 0, screen_x - 1);
	}

	wresize(stat_win, 1, screen_x);
	mvwin(stat_win, screen_y - 2, 0);
	wresize(status_bar, 1, screen_x - 19);

#ifdef ENABLE_EXTENDED_KEYS
	/* For FreeBSD */
	keypad(status_bar, TRUE);
#endif /* ENABLE_EXTENDED_KEYS */

	mvwin(status_bar, screen_y - 1, 0);
	wresize(pos_win, 1, 13);
	mvwin(pos_win, screen_y - 1, screen_x - 13);

	wresize(input_win, 1, 6);
	mvwin(input_win, screen_y - 1, screen_x - 19);

	curs_set(0);
}

void
redraw_window(void)
{
	resize_window();

	if(curr_stats.show_full)
	{
		redraw_full_file_properties(curr_view);
		curr_stats.freeze = 0;
		curr_stats.need_redraw = 0;
		return;
	}

	change_directory(curr_view, curr_view->curr_dir);
	load_dir_list(curr_view, 1);

	if(curr_stats.view)
	{
		wclear(other_view->win);
		quick_view_file(curr_view);
	}
	else
	{
		change_directory(other_view, other_view->curr_dir);
		load_dir_list(other_view, 1);
	}

	update_stat_window(curr_view);
	update_pos_window(curr_view);

	if(curr_view->selected_files)
	{
		char status_buf[24];
		snprintf(status_buf, sizeof(status_buf), "%d %s Selected",
				curr_view->selected_files,
				curr_view->selected_files == 1 ? "File" : "Files");
		status_bar_message(status_buf);
	}
	else
		clean_status_bar();

	update_pos_window(curr_view);

	update_all_windows();

	moveto_list_pos(curr_view, curr_view->list_pos);
	wrefresh(curr_view->win);
	curr_stats.freeze = 0;
	curr_stats.need_redraw = 0;

	if(curr_stats.errmsg_shown)
	{
		redraw_error_msg(NULL, NULL);
		redrawwin(error_win);
		wnoutrefresh(error_win);
		doupdate();
	}
}

void
clean_status_bar()
{
	werase(status_bar);
	wnoutrefresh(status_bar);
}

static void
switch_views(void)
{
	FileView *tmp = curr_view;
	curr_view = other_view;
	other_view = tmp;

	load_local_options(curr_view);
}

void
change_window(void)
{
	switch_views();

	if(curr_stats.number_of_windows != 1)
	{
		wattroff(other_view->title, A_BOLD);
		wattroff(other_view->win,
				COLOR_PAIR(other_view->color_scheme + CURR_LINE_COLOR) | A_BOLD);
		mvwaddstr(other_view->win, other_view->curr_line, 0, "*");
		erase_current_line_bar(other_view);
		werase(other_view->title);
		wprintw(other_view->title, "%s", replace_home_part(other_view->curr_dir));
		wnoutrefresh(other_view->title);
	}

	if(curr_stats.view)
	{
		wbkgdset(curr_view->title,
				COLOR_PAIR(BORDER_COLOR + curr_view->color_scheme));
		wbkgdset(curr_view->win,
				COLOR_PAIR(WIN_COLOR + curr_view->color_scheme));
		change_directory(other_view, other_view->curr_dir);
		load_dir_list(other_view, 1);
		change_directory(curr_view, curr_view->curr_dir);
		load_dir_list(curr_view, 1);
	}

	wattron(curr_view->title, A_BOLD);
	werase(curr_view->title);
	wprintw(curr_view->title, "%s", replace_home_part(curr_view->curr_dir));
	wnoutrefresh(curr_view->title);

	wnoutrefresh(other_view->win);
	wnoutrefresh(curr_view->win);

	change_directory(curr_view, curr_view->curr_dir);

	if(curr_stats.number_of_windows == 1)
		load_dir_list(curr_view, 1);

	moveto_list_pos(curr_view, curr_view->list_pos);
	werase(status_bar);
	wnoutrefresh(status_bar);

	if(curr_stats.number_of_windows == 1)
		update_all_windows();
}

static void
update_view(FileView *win)
{
	touchwin(win->title);
	touchwin(win->win);

	redrawwin(win->title);
	redrawwin(win->win);

	wnoutrefresh(win->title);
	wnoutrefresh(win->win);
}

void
update_all_windows(void)
{
	touchwin(lborder);
	touchwin(stat_win);
	touchwin(status_bar);
	touchwin(pos_win);
	touchwin(input_win);
	touchwin(rborder);

	/*
	 * redrawwin() shouldn't be needed.  But without it there is a
	 * lot of flickering when redrawing the windows?
	 */

	redrawwin(lborder);
	redrawwin(stat_win);
	redrawwin(status_bar);
	redrawwin(pos_win);
	redrawwin(input_win);
	redrawwin(rborder);

	/* In One window view */
	if(curr_stats.number_of_windows == 1)
	{
		update_view(curr_view);
	}
	/* Two Pane View */
	else
	{
		touchwin(mborder);
		redrawwin(mborder);
		wnoutrefresh(mborder);

		update_view(&lwin);
		update_view(&rwin);
	}

	wnoutrefresh(lborder);
	wnoutrefresh(stat_win);
	wnoutrefresh(status_bar);
	wnoutrefresh(pos_win);
	wnoutrefresh(input_win);
	wnoutrefresh(rborder);

	if(!curr_stats.errmsg_shown && curr_stats.vifm_started == 2)
		doupdate();
}

void
update_input_bar(wchar_t c)
{
	wchar_t buf[] = {c, '\0'};

	if(!curr_stats.use_input_bar)
		return;

	if(getcurx(input_win) == getmaxx(input_win) - 1)
	{
		mvwdelch(input_win, 0, 0);
		wmove(input_win, 0, getmaxx(input_win) - 2);
	}

	waddwstr(input_win, buf);
	wrefresh(input_win);
}

void
clear_num_window(void)
{
	if(curr_stats.use_input_bar)
	{
		werase(input_win);
		wrefresh(input_win);
	}
}

/* msg can't be NULL
 * period - how often status bar should be updated
 * if period equals 0 reset inner counter
 */
void
show_progress(const char *msg, int period)
{
	static char marks[] = {'|', '/', '-', '\\'};
	static int count = 0;
	static int pause = 1;

	char buf[strlen(msg) + 2 + 1];

	if(period == 0)
	{
		pause = 1;
		return;
	}

	pause++;

	if(pause%period != 0)
		return;

	pause = 1;

	sprintf(buf, "%s %c", msg, marks[count]);
	status_bar_message(buf);
	wrefresh(status_bar);

	count = (count + 1) % sizeof(marks);
}

void
redraw_lists(void)
{
	draw_dir_list(&lwin, lwin.top_line);
	draw_dir_list(&rwin, rwin.top_line);
	moveto_list_pos(curr_view, curr_view->list_pos);
}

void
load_color_scheme(const char *name)
{
	int i;
	int color_scheme;

	i = find_color_scheme(name);
	if(i < 0)
	{
		show_error_msg("Color Scheme", "Invalid color scheme name");
		return;
	}

	cfg.color_scheme_cur = i;
	cfg.color_scheme = MAXNUM_COLOR*cfg.color_scheme_cur;
	color_scheme = i*MAXNUM_COLOR;

	wbkgdset(lborder, COLOR_PAIR(color_scheme + BORDER_COLOR));
	werase(lborder);
	wbkgdset(mborder, COLOR_PAIR(color_scheme + BORDER_COLOR));
	werase(mborder);
	wbkgdset(rborder, COLOR_PAIR(color_scheme + BORDER_COLOR));
	werase(rborder);
	wbkgdset(stat_win, COLOR_PAIR(color_scheme + BORDER_COLOR));

	wbkgdset(menu_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	wbkgdset(sort_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	wbkgdset(change_win, COLOR_PAIR(color_scheme + WIN_COLOR));
	wbkgdset(error_win, COLOR_PAIR(color_scheme + WIN_COLOR));

	draw_dir_list(curr_view, curr_view->top_line);
	moveto_list_pos(curr_view, curr_view->list_pos);
	draw_dir_list(other_view, other_view->top_line);
}

/* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
